/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "streaming_conditioner.hh"
#include "wtile.hh"
#include "matlab_fcs.hh"
#include "medianmeanaveragespectrum.hh"
#include "spline_psd.hh"

#include "DVecType.hh"
#include "TSeries.hh"
#include "fSeries/ASD.hh"
#include "fSeries/PSD.hh"
#include "FIRdft.hh"
#include "FIRdesign.hh"

//---------------------
#include <iostream>
using namespace std;
//---------------------

using namespace wpipe;
using namespace containers;

// $Id: streaming_conditioner.m 3070 2019-08-10 21:38:01Z jzweizig $

//  WhitenedDft results verified 2011.06.28
//  non-standard? normalization from whitened dft? Maybe an extra factor in 
//  whitening coefficients.


//======================================  Condition constructor
streaming_conditioner::streaming_conditioner(const wtile& tiling, 
					     bool doubleWhiten,
					     int dbugLevel)
   : whiten_stat(wNone), debugLevel(dbugLevel)
{
   //-----------------------------------  validate tiling structure
   if (tiling.id() != "Discrete Q-transform tile structure") {
      error("input argument is not a discrete Q transform tiling structure");
   }

   //-----------------------------------  set up the resample
   fSample    = tiling.sampleFrequency();
   resample.set_sample_rate(fSample);

   //-----------------------------------  determine required data lengths
   duration   = tiling.duration();
   dataLength = size_t(fSample * double(duration));
   if (tiling.whiteningDuration() > 0) {
      whitener.set_stride(tiling.whiteningDuration());
      whiten_stat = doubleWhiten ? wDouble : wSingle;
   }

   //-----------------------------------  Design high-pass filter if requested
   highPassCutoff = tiling.highPassCutoff();
   if (highPassCutoff > 0) {
      double Hz = fSample;
      double f0 = highPassCutoff;
      double df = 0.25*f0;
      double attn = 100;
      if (debugLevel) {
	 cout << "streaming_conditioner: Highpass filter design parameters: "
	      << "f0: " << f0 << "Hz attn: " << attn << "dB df: " << df
	      << "Hz" << endl;
      }
      
      //--------------------------------  Calculate the length
      int Nf = int((attn-7.95)/(14.36*2*df/Hz)) + 1;
      int nsFact = 2 * Hz / gcd(1000000000, Hz);
      if ((Nf-1) % nsFact != 0) {
	 Nf = ((Nf - 1) / nsFact + 1) * nsFact + 1;
      }
      FIRdft hpf(dFirW(Nf, Hz, "Kaiser", "HighPass", f0, 0.0, attn, df));
      hpf.setMode(FIRdft::fm_zero_phase);
      
      if (debugLevel) {
	 cout << "streaming_conditioner: Highpass filter designed. delay ="
	      << hpf.getTimeDelay() << endl;
      }
      highPassFilter.set(hpf);
   }
   
   //----------------------------  Reset all channel vectors
   reset();
}

//======================================  Condition destructor
streaming_conditioner::~streaming_conditioner(void) {
}

//======================================  Multi channel conditioning
int
streaming_conditioner::condition(const TSeries& rawData, const Time& t0,
				 bool valid) 
{
   if (debugLevel > 1) {
      cout << "Enter streaming_conditioner::condition, t0 = " << t0 << endl;
      rawData.dump_header(cout);
   }
   
   //-----------------------------------  check for valid input data
   if (rawData.getStartTime() != current && current.ne_0()) {
      cerr << "streaming_conditioner: phase error. current = " << current
	   << " raw data start = " << rawData.getStartTime() << endl;
      throw runtime_error("stream_conditioner: raw data out of phase");
   }

   //-----------------------------------  resample
   TSeries resData = resample(rawData);
   if (debugLevel > 1) {
      cout << "streaming_conditioner: resampled data" << endl;
      resData.dump_header(cout);
   }
   
   //-----------------------------------  Save raw (resampled) data
   int errflag = raw.Append(resData);
   if (errflag) error("streaming_conditioner: resampled data out of phase");
   if (t0.ne_0() && raw.getStartTime() < t0) {
      raw.eraseStart(t0 - raw.getStartTime());
   } else if (!t0 && raw.getInterval() > duration) {
      raw.eraseStart(raw.getInterval() - duration);
   }

   //-----------------------------------  High-pass filtering
   TSeries hpIncrement;
   if (highPassCutoff > 0) {
      if (debugLevel) cout << "    high-pass filter data, fmin=" 
			   << highPassCutoff << endl;
      
      // -------------------------------   Apply zero-phase high-pass filter
      hpIncrement = highPassFilter(resData);

      // include high pass filter in conditioning coefficients
      static_cast<fSeries&>(coefficients) = hpfResponse;
   }

   //-----------------------------------  No hp filtering requested, do nothing
   else {
      hpIncrement = resData;
   }  // end test for high pass filtering

   errflag = highPassData.Append(hpIncrement);
   if (errflag) error("stream_conditioner: high-passed data out of phase");
   
   if (debugLevel > 1) {
      cout << "streaming_conditioner: high-passed data" << endl;
      highPassData.dump_header(cout);
   }
   
   current = rawData.getEndTime();
   if (!valid || !t0) {
      if (highPassData.getInterval() > duration) {
	 highPassData.eraseStart(highPassData.getInterval() - duration);
      }
      return 0;
   }
   
   //-----------------------------------  Toss extra data. test for full block
   Interval t0_offset = t0 - highPassData.getStartTime();
   if (t0_offset < Interval(0)) return 2; //----  No data at block start (t0).
   if (t0_offset > Interval(0)) highPassData.eraseStart(t0_offset);
   if (highPassData.getInterval() < duration) return 1; 
   
   //-----------------------------------  Calculate DFT of filtered data
   TSeries hpSegment(highPassData.extract(t0, duration));
   highPassDft = DFT(hpSegment);
 
   //-----------------------------------  Whiten the data if requested
   if (whiten_stat == wNone) {
      wlog(debugLevel, 1, "   skip whitening");
      whitenedDft = highPassDft;
   }
   
   //-----------------------------------  whitening is requested
   else {
      wlog(debugLevel, 1, "    whiten data");

      if (whitener.currentTime() != hpIncrement.getStartTime() && 
	  whitener.currentTime().ne_0()) {
	 whitener.reset();
      }

      if (debugLevel > 1) {
	 cout << "streaming_conditioner: whiten segment: " << endl;
	 hpIncrement.dump_header(cout);
      }
      whitener.add(hpIncrement);
      PSD PSDintermediate = whitener.get_psd();
      if (debugLevel > 1) {
	 cout << "streaming_conditioner: whitening output psd: " << endl;
	 PSDintermediate.dump_header(cout);
      }
						      
      // -- Correct normalization:
      //    Note the factor of 2 because highPassDft is already folded.
      PSDintermediate *= fSample * 2 * (highPassDft.size() - 1);
      
      // -- interpolate to finer resolution
      PSDintermediate = spline_psd(PSDintermediate, 1.0/double(duration));

      // make theCoefficients directly from the power spectrum
      fSeries theCoefficients(PSDintermediate);
      size_t N = theCoefficients.size();
      DVectD& dvd = dynamic_cast<DVectD&>(theCoefficients.refDVect());
      for (size_t i=0; i<N; ++i) {
	 if (dvd[i] != 0) dvd[i] = sqrt(2.0/dvd[i]);
      }

      if (highPassCutoff > 0) {
	 int hpBin= theCoefficients.getBin(highPassCutoff);
	 theCoefficients.refDVect().replace_with_zeros(0, hpBin, hpBin);
      }

      // include whitening filter in conditioning coefficients
      if (!coefficients.empty()) {
	 coefficients *= theCoefficients;
      } else {	
	 static_cast<fSeries&>(coefficients) = theCoefficients;
      }
      
      // apply whitening filter
      whitenedDft  = highPassDft;
      whitenedDft *= theCoefficients;

      // reapply whitening filter if double whitening requested
      if (whiten_stat == wDouble) {
	 wlog(debugLevel, 1, "    double whiten data");
	 whitenedDft *= theCoefficients;
	 coefficients *= theCoefficients;
      }
   }
   return 0;
}

//======================================  Get coefficients
void
streaming_conditioner::coefficientDFT(containers::DFT& dft) const {
   dft = coefficients;
}

//======================================  Calculate time delay
Interval
streaming_conditioner::getTimeDelay(void) const {
   Interval delay = resample.getTimeDelay();
   if (!highPassFilter.null()) delay += highPassFilter->getTimeDelay();
   return delay;
}

//======================================  Get a raw DFT
const containers::DFT&
streaming_conditioner::rawDFT(void) {
   if (rawDft.empty()) {
      rawDft = DFT(raw);
   }
   return rawDft;
}

//=======================================  Reset the conditioner
void 
streaming_conditioner::reset(void) {
   if (debugLevel) {
      cout << "Enter streaming_conditioner::reset()" << endl;
   }

   //------------------------------------  Clear and reset all the data vectors
   current = Time(0);
   resample.reset();
   raw.Clear(); 
   highPassData.Clear();
   highPassDft.clear();
   whitenedDft.clear();
   coefficients.clear();
   highPassFilter.reset();
   whitener.reset();
}

//=======================================  Get high-passed data
TSeries 
streaming_conditioner::highPassedData(void) const {
   return highPassData;
}

//=======================================  Get whitened data
TSeries 
streaming_conditioner::whitenedData(void) const {
   TSeries temp_ts = whitenedDft.iFFT();
   temp_ts.setUnits("whitened");
   return temp_ts;
}

void 
streaming_conditioner::whitenedData(TSeries& ts) const {
   ts = whitenedDft.iFFT();
   ts.setUnits("whitened");
}

void
streaming_conditioner::whitenedDFT(containers::DFT& dft) const {
   dft = whitenedDft;
}
